// LIBRARIES

#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "SDLmain.lib")
#pragma comment(lib, "zdll.lib")

// HEADER FILES


#include <stdlib.h>
#include <string.h>
#include <Windows.h>
#include <signal.h>

#include "SDL.h"
#include "zlib.h"

// GLOBAL VARIABLES

bool		terminate_app	= false;		// Set to TRUE to signify end of all
int			psg_period[4]	= {0,};			// Periods of the 3 channels
int			psg_level[4]	= {15,15,15,15};// Level of the PSG channels
int			psg_status[4]	= {1,1,1,1};	// High = 1, Low = 0
int			psg_timeout[4]	= {0,};			// Timeout for the PSG channels
int			delay			= 0;			// Delay before next command
int*		vgm_data;						// Expanded VGM data
DWORD64		vgm_data_ptr	= 0;			// Where are we? (Progress)
int			last_chan		= 0;			// Last channel we wrote to
bool		white_noise		= true;			// Periodic/White noise?
int			play_mode		= 0;			// 0=stopped, 1=playing, 2=paused.
int			freq_scaler_den	= 0;			// Amount to scale frequency by
int			freq_scaler_num	= 0;			// Amount to scale frequency by
int			psg_pan[4][2];					// Pan on each side
int			noise_freq		= 0;			// Noise channel frequency
int			lfsr_noise		= 0;			// Linear feedback shift register
int			noise_mode		= 0x9;			// Noise mode for hardware (0x9=SMS/GG, 0x3=SC-3000H, 0x8005=BBC Micro)
DWORD64		samples_played	= 0;			// Number of samples played
HANDLE		r_hnd			= NULL;			// Read handle from our console window
HANDLE		w_hnd			= NULL;			// Write handle from our console window
int			sound_mode		= 0;			// Which noise to use?
int			modes[3]		= {0x9,0x6,0x8005};
char*		mode_names[3]	= {"SMS","SC3","BBC"};
bool		run_loop		= true;			// Loop the song at loop points?
int			edge_clock		= 0;			// So the noise register only updates once every 2 clocks
int			old_sound_mode	= -1;			// What was the sound mode of the last frame?
int			enabled[4]		= {1,1,1,1};	// Which channels are enabled?
int			antialias[4][2]	= {{0,},{0,}};	// Counters to antialias the sound		
int			aa_samples		= 0;			// How many samples were taken for the antialias?
// VGM FILE DETAILS

int			file_end_offset	= 0;			// Pointer to end of file
int			vgm_version_no	= 0;			// Version number
int			psg_speed		= 0;			// Clock rate of PSG
int			fm_speed		= 0;			// Clock rate of FM chip
int			gd3_tag_offset	= 0;			// Offset to GD3 tag
DWORD64		sample_count	= 0;			// Number of samples in file count
DWORD64		loop_offset		= 0;			// Offset of loop position
DWORD64		loop_samples	= 0;			// Number of samples in loop
int			record_rate		= 0;			// 50/60Hz
int			gd3_version		= 0;			// GD3 tag version
int			gd3_size		= 0;			// GD3 tag size


// FUNCTION TO CLEAN UP

void clean_up() {
	SDL_CloseAudio();
	SDL_Quit();
	if (vgm_data) delete vgm_data;	
}

// CLOSE HANDLER

BOOL ctrl_handler(DWORD ctrl_type) {
	clean_up();
	terminate_app = true;
	return false;
}

void exit_handler(int sig) {
	ctrl_handler(CTRL_CLOSE_EVENT);
}


// SOUND [PSG EMULATION]

// Get the next command from the VGM file, or return 0x66 (end of song) if
// we've gone past the end of the buffer.

int get_next_command() {
	if (vgm_data_ptr > file_end_offset-0x44) {
		return 0x66;
	} else {
		return vgm_data[vgm_data_ptr++];
	}
}

// Function to calculate parity bit (used with shift register)
// Thanks to Dave (finaldave) [taken from Maxim's SN79489 notes]

int parity(int val) {
	val^=val>>8;
	val^=val>>4;
	val^=val>>2;
	val^=val>>1;
	return val&1;
}

void play_sample() {

	++samples_played;

	if (delay==0) {
		// Update...
		int cmd = get_next_command();

		if (cmd==0x50) {
			// It's a PSG write
			int psg_write = get_next_command();
			
			// Does it latch a channel?
			if (psg_write & 0x80) {
				last_chan = (psg_write>>5)&3;
				int psg_data = psg_write & 0xF;
				// Right, now, is it going to affect volume or period?
				if (psg_write & 0x10) {
					// Volume
					psg_level[last_chan] = 0xF-psg_data;
				} else {
					// Period
					if (last_chan==3) {
						if (psg_write & 4) {
							white_noise = true;
						} else {
							white_noise = false; 
						}
						lfsr_noise = 0x8000; // Reset noise frequency
						noise_freq = psg_write & 3;
					} else {
						psg_period[last_chan]&=0xFFF0;
						psg_period[last_chan]+=psg_data;
					}
				}
			} else {
				// Just update:
				if (last_chan!=3) {
					psg_period[last_chan]&=0xF;
					psg_period[last_chan]+=(psg_write&0x3F)<<4;
				}
			}
		} else if (cmd==0x61) {
			delay  = get_next_command();
			delay += get_next_command()<<8;
		} else if (cmd==0x62) {
			delay = 735;
		} else if (cmd==0x63) {
			delay = 882;
		} else if (cmd==0x66) {
			if ((!run_loop||(loop_offset==0&&loop_samples==0)) || loop_offset>=sample_count) {
				play_mode = 0;
				samples_played = 0;
				vgm_data_ptr = 0;
				SDL_PauseAudio(true);
			} else {
				vgm_data_ptr = loop_offset-0x1C;
				samples_played-=loop_samples;
			}
		} else if (cmd==0x4F) {
			// Handle stereo
			int stereo_write = get_next_command();
			for (int c=1; c>=0; --c) {
				for (int j=0; j<4; ++j) {
					psg_pan[j][c] = (stereo_write&1);
					stereo_write>>=1;
				}
			}
		} else if ((cmd&0xF0)==0x50) {
			get_next_command();
			get_next_command();
		}
	} else {
		--delay;
	}


	if (play_mode!=1) return;

	// Update output levels:
	
	// Reset totals counter
	for (int j=0; j<4; ++j) {
		for (int c=0; c<2; ++c) {
			antialias[j][c] = 0;
		}
	}
	aa_samples = 0;
	bool generate_squares = true;
	while (generate_squares) {
		for (int j=0; j<4; ++j) {
			if (--psg_timeout[j]<=0) {
				if (j==3) {
					// Handle shift register
					switch (noise_freq) {
						case 0:
							psg_timeout[3] = 0x10;
							break;
						case 1:
							psg_timeout[3] = 0x20;
							break;
						case 2:
							psg_timeout[3] = 0x40;
							break;
						case 3:
							psg_timeout[3] =  psg_period[2];
							break;
					}
					edge_clock = 1-edge_clock;
					if (edge_clock) {
						// lsfr_noise equation lifted from Maxim's SN79489 notes.
						lfsr_noise=(lfsr_noise>>1)|((white_noise?parity(lfsr_noise&noise_mode):lfsr_noise&1)<<15);
						psg_status[3] = lfsr_noise&1;
					}
				} else {
					if (psg_period[j]) {
						psg_status[j]=1-psg_status[j];
						psg_timeout[j]=psg_period[j];
					}
				} // end if (j==3)
			} // PSG timeout hit 0
			for (int c=0; c<2; ++c) {
				antialias[j][c]+=psg_level[j]*psg_pan[j][c]*psg_status[j]*enabled[j];
			}
		}  // for j=0 to 4
		++aa_samples;
		freq_scaler_num+=44100;
		if (freq_scaler_num>freq_scaler_den) {
			freq_scaler_num-=freq_scaler_den;
			generate_squares = false;
		}
	}// end of while loop
}


// CALLBACK FUNCTION TO UPDATE SOUND OUTPUT

void update_sound(void *userdata, Uint8 *stream, int len) {


	// We're going to update and play some sound...

	for (int i=0; i<len; i+=4) {

		if (play_mode==1) play_sample();

		// Stereo mixer
		for (int c=0; c<2; ++c) {

			int output_level = 0;
			
			for (int j=0; j<4; ++j) {
				int out_v = (antialias[j][c]<<9)/aa_samples;
				output_level += out_v;
			}

			// Clip (not very accurate or neat, but at least it stops overflow from buzzing)
			if (output_level>0xFFFF) output_level=0xFFFF;
			if (output_level<0x0000) output_level=0x0000;
		
			*stream++ =   output_level    &0xFF;
			*stream++ =  (output_level>>8)&0xFF;
		
		}
	}

}

// MAIN ENTRY POINT

int main(int argc, char* argv[]) {

	printf("VGM Player by Ben Ryves version 1.1 [September 2005]\n\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\n\n");
	
	if (argc==1) {
		printf("Usage: VGMPLAY <filename>\n");
		return 1;
	}

	// Set up close handlers.
	// signal(SIGTERM,...) appears to do nothing on its own,
	// SetConsoleCtrlHandler() only seems to work every other time.
	// Use both together and it all seems to work hunky-dory
	SetConsoleCtrlHandler((PHANDLER_ROUTINE)ctrl_handler,true);
	signal(SIGTERM,exit_handler);

	// Set up our console window:
	r_hnd = GetStdHandle(STD_INPUT_HANDLE);
	w_hnd = GetStdHandle(STD_OUTPUT_HANDLE);

	// Try to open the VGM file:
	gzFile vgm_file = gzopen(argv[1],"rb");
	if (!vgm_file) {
		printf("Failed!\n");
		return 1;
	} else {
		// At this point, try to read the file:
		char detect_header[5] = {0,};
		gzread(vgm_file,detect_header,4);
		if (strcmp(detect_header,"Vgm ")) {
			printf("Not a valid VGM file.\n");
			return 1;
		} else {
			gzread(vgm_file,&file_end_offset,4);
			gzread(vgm_file,&vgm_version_no,	4);
			gzread(vgm_file,&psg_speed,		4);
			if (psg_speed==0) {
				printf("This player only supports PSG (SN76496) VGM files.\n");
				return 1;
			} else {
				freq_scaler_num = psg_speed/8;
				freq_scaler_den = freq_scaler_num/2;
				gzread(vgm_file,&fm_speed,		4);
				gzread(vgm_file,&gd3_tag_offset,4);
				gzread(vgm_file,&sample_count,	4);
				gzread(vgm_file,&loop_offset,	4);
				gzread(vgm_file,&loop_samples,	4);
				gzread(vgm_file,&record_rate,	4);

				gzseek(vgm_file,0x40,0);

				int total_data_size = file_end_offset-0x44;
				vgm_data = new int[total_data_size];
				for (int j=0; j<total_data_size; ++j) {
					unsigned char cmd_byte = 0;
					gzread(vgm_file,&cmd_byte,1);
					vgm_data[j]=cmd_byte;
				}
				// Display the GD3 data, if available:
				if (gd3_tag_offset==0) {
					printf("No GD3 tag available.\n");
				} else {

					// Jump to GD3 tag

					gzseek(vgm_file,gd3_tag_offset+0x14,0);
					char detect_header[5] = {0,};
					gzread(vgm_file,detect_header,4);
					if (strcmp(detect_header,"Gd3 ")!=0) {
						printf("Invalid GD3 tag.\n");
					} else {
						
						unsigned short info[1024];
						
						gzread(vgm_file,&gd3_version,4);
						gzread(vgm_file,&gd3_size,4);

						int gd3_size_count = gd3_size;
						int tag_type = 0;

						while (gd3_size>0) {
							for (int i=0; i<1024; ++i) {
								gzread(vgm_file,&info[i],2);
								gd3_size-=2;
								if (info[i]==0) {
									i=1024;
								}
							}
							// Is it a system name?
							if (tag_type==4) {
								// See if we can preset a sound type
								if (!wcscmp(L"Sega SC-3000",info)) {
									sound_mode = 1;
								} else if (!wcscmp(L"BBC Micro",info) || !wcscmp(L"Acorn BBC Micro",info)) {
									sound_mode = 2;
								}
								noise_mode = modes[sound_mode];
							}

							// Display the tag information
							if (wcslen(info) && (tag_type!=1&&tag_type!=3&&tag_type!=5&&tag_type!=7)) {
								switch (tag_type) {
									case 0: printf("Track:\t"); break;
									case 1: printf("Track:\t"); break;
									case 2: printf("Game:\t"); break;
									case 3: printf("Game:\t"); break;
									case 4: printf("System:\t"); break;
									case 5: printf("System:\t"); break;
									case 6: printf("Author:\t"); break;
									case 7: printf("Author:\t"); break;
									case 8: printf("Date:\t"); break;
									case 9: printf("Ripper:\t"); break;
									case 10: printf("Notes:\t"); break;
									default: printf("Misc:\t");
								}
								wprintf(info);
								printf("\n");
							}
							++tag_type;
						}					
					}
				}
			}			
		}
		gzclose(vgm_file);
	}

	// Initialise SDL:
	if (SDL_Init(SDL_INIT_AUDIO)<0)	{
		printf("SDL Audio initialization failed: %s\n",SDL_GetError());
		SDL_Quit();
		return 1;
	}
	// Try to open the audio device:

	SDL_AudioSpec desired;
	SDL_AudioSpec sound_spec;
	desired.freq		= 44100;
	desired.format		= AUDIO_U16;
	desired.channels	= 2;
	desired.samples		= 4096;
	desired.callback	= update_sound;
	desired.userdata	= NULL;

	if (SDL_OpenAudio(&desired, &sound_spec) < 0){
		printf("Opening audio device failed!\n");
		SDL_Quit();
		return 1;
	}
	// Start playing:

	vgm_data_ptr = 0;
	play_mode = 1;
	for (int i=0; i<4; ++i) for (int j=0; j<2; ++j) {
		psg_pan[i][j] = 1;
	}
	SDL_PauseAudio(false);

	printf("\n");

	while (!terminate_app) {

		// Handle status changes:
		if (play_mode==0) {
			vgm_data_ptr = 0;
		}

		// Handle input events:

		DWORD num_events = 0;
		DWORD num_events_read = 0;
		GetNumberOfConsoleInputEvents(r_hnd, &num_events);
		if (num_events!=0) {
			INPUT_RECORD *event_buffer = new INPUT_RECORD[num_events];
			ReadConsoleInput(r_hnd, event_buffer, num_events, &num_events_read);
			for (DWORD i=0; i<num_events_read; ++i) {
				if (event_buffer[i].EventType==KEY_EVENT && event_buffer[i].Event.KeyEvent.bKeyDown) {
					WORD key = event_buffer[i].Event.KeyEvent.wVirtualKeyCode;
					if (key==VK_ESCAPE) {
						terminate_app = true;
					} else if (key==VK_SPACE) {
						if (play_mode==1) {
							play_mode = 2;
						} else {
							play_mode = 1;
						}
					} else if (key==190) {
						play_mode = 0;
					} else if (key==78) {
						sound_mode+=1;
						sound_mode%=3;
						noise_mode = modes[sound_mode];						
					} else if (key==76) {
						run_loop = !run_loop;
					} else if (key>48&&key<53) {
                        enabled[key-49] = 1-enabled[key-49];						
					}
				} else if (event_buffer[i].EventType==MOUSE_EVENT && (event_buffer[i].Event.MouseEvent.dwButtonState & FROM_LEFT_1ST_BUTTON_PRESSED)) {
					COORD mouse_pos = event_buffer[i].Event.MouseEvent.dwMousePosition;
					// Which line is our cursor on?
					CONSOLE_SCREEN_BUFFER_INFO cursor_pos;
					GetConsoleScreenBufferInfo(w_hnd, &cursor_pos);
					
					if (mouse_pos.Y==cursor_pos.dwCursorPosition.Y) {

						if (mouse_pos.X<32) {
			
							SDL_PauseAudio(true);
							play_mode = 0;
							DWORD64 sample_pos = ((unsigned int)mouse_pos.X*sample_count)/32;

							if (sample_pos<samples_played) {
								vgm_data_ptr = 0;
								samples_played = 0;
							}
							
							while (samples_played<sample_pos && samples_played<sample_count) {
								DWORD64 old_sample_pos = samples_played;
								play_sample();
								if (samples_played<=old_sample_pos) break;
							}
							play_mode = 1;
							SDL_PauseAudio(false);

						} else {
							if (mouse_pos.X>32&&mouse_pos.X<36) {
								if (play_mode==1) {
									play_mode = 2;
								} else {
									play_mode = 1;
								}
							} else if (mouse_pos.X>35&&mouse_pos.X<39) {
								play_mode = 0;
							} else if (mouse_pos.X>38&&mouse_pos.X<42 || mouse_pos.X>48&&mouse_pos.X<52) {
								sound_mode+=1;
								sound_mode%=3;
								noise_mode = modes[sound_mode];
							} else if (mouse_pos.X>41&&mouse_pos.X<45) {
								run_loop = !run_loop;
							} else if (mouse_pos.X>44&&mouse_pos.X<48) {
								terminate_app = true;
							} else if (mouse_pos.X>53 && mouse_pos.X<58) {
								enabled[mouse_pos.X-54] = 1-enabled[mouse_pos.X-54];
							}
						}
					}
                }
			}
			delete event_buffer;
		}

		if (play_mode==1) {
			SDL_PauseAudio(false);
		} else {
			SDL_PauseAudio(true);
		}

		// Status:

		char progress[33] = {0,};
		DWORD64 tot_progress = (samples_played*32)/(sample_count);
		if (tot_progress>31) tot_progress = 31;

		for (int i=0; i<32; ++i) {
			if (i<tot_progress) {
				progress[i]=(char)0xB2;
			} else if (i>tot_progress) {
				progress[i]=(char)0xB0;
			} else {
				progress[i]=(char)0xDB;
			}
		}

		int minutes = 0;
		
		DWORD64 hundred_total_time = samples_played/441;
		float time_played = (float)hundred_total_time/100;

		if (time_played<0) time_played = 0;
		while (time_played>60) {
			minutes+=1;
			time_played-=60;
		}
		if (time_played<0) time_played = 0;
        
	
		printf(progress);
		printf(" [");
		if (play_mode==1) {
			printf("\x1F");
		} else {
			printf("\x10");
		}
		printf("][\xFE][\x0E][");
		if (run_loop) {
			printf("\x1A");
		} else {
			printf("\x1D");
		}
		printf("][x] %s [",mode_names[sound_mode]);
	
		for (int j=0; j<4; ++j) {
			if (enabled[j]) {
				printf("*");
			} else {
				printf("\xFA");
			}
		}

		printf("] (%d:",minutes);
		if (time_played<10) printf("0");
		printf("%.2f)        \r",time_played);


		// Handle status changes:
		if (play_mode==0) {
			vgm_data_ptr = 0;
			samples_played = 0;
		}

		// Idle...
		SDL_Delay(50);
	}

	clean_up();
	return 0;

}